/**************************************************************************
 * camera.c
 * Description: This file includes any fucntions necessary to interact
 *              with the serial camera for the UR Smart Home.
 *
 * Functions: For more information please see the camera.h file in the
 *     		include folder.
 * Written by: Hector Nuez, Sina Karimi, Ryan Campbell
 * Date: February 23, 2010
 ***************************************************************************/

#include "camera.h"

Camera::Camera()
{
    pioCtrl_desc.pio_base = PIOA_BASE;
    pioCtrl_desc.periph_id = PIOA_ID;
    pioCtrl_desc.pio_number = NB_PIOA;

    comPort.usart_base = USART2_BASE;
    comPort.pio_ctrl = &pioCtrl_desc;
    comPort.pin_rxd = PIORXD2;
    comPort.pin_txd = PIOTXD2;
    comPort.pin_sck = PIOSCK2;
    comPort.periph_id = US2_ID;
    Init();
}

/*********************************** camera_init **********************************/
void Camera::Init()
{
    //Initializing the Serial Port (USART) named "comPort"
    //The following has been taken from m55800_lib16\periph\usart\usart.h file
    //US_ASYNC_MODE = Standard Asynchronous Mode : 8 bits , 1 stop , no parity
    //And these ones where decalred in camera.h file
    //Choose 14400 bps baud rate for now and expeiment later with higher speeds
    //timeguard is disabled because tramission of commands are expected
    //  to be continuous
    at91_usart_open (&comPort,US_ASYNC_MODE,BAUDRATE_57600,TIME_GUARD_DISABLE);

    /*create the semaphore */
    CameraSem = OSSemCreate(1);
}

int Camera::powerOn()
{
    return send_SYNC() && send_INITIALIZE() && send_SET_PACKG_SIZE();
}

/*********************************** is_ack ***************************/    

int Camera::is_ack(char * rxBuffer, u_int commandID)
{
    return ((u_int)rxBuffer[0] == TRANSMIT)&&
	    ((u_int)rxBuffer[1] == ACK) &&
	    ((u_int)rxBuffer[2] == commandID) ? TRUE : FALSE;
}

/*********************************** is_sync ***************************/  
int Camera::is_sync(char * rxBuffer)
{
    return ((u_int)rxBuffer[0] == TRANSMIT) &&
	((u_int)rxBuffer[1] == SYNC) ? TRUE : FALSE;
}


/*********************************** get_picture ***************************/  

int Camera::get_picture(char * filePath)
{
    char rxBuffer[DATA_PACKG_SIZE + DATA_PACKG_OVERHEAD];
    //char picBuffer[20000]; make this a data member
    
    char * pPicBuffer = picBuffer;
    char * pImageBuffer; 

    int dataLength;
    u_int totPackgNum;
    u_int dataSize = 0;
    u_int checkSum = 0;

    int i = 0;
    int j = 0;

    char myPackgIdNum[2];

    memset(picBuffer, 0, 10000);
    memset(rxBuffer, 0, DATA_PACKG_SIZE + DATA_PACKG_OVERHEAD);

    

    if (!send_GET_PICTURE(&dataLength))
	return FALSE;
    
    totPackgNum = dataLength/(DATA_PACKG_SIZE - DATA_PACKG_HEADER) + 1;
#ifdef DEBUG
    printf("totPackgNum = %d\n", totPackgNum);

#endif

    //concatenate image data into picBuffer	
    for(i = 0; i < totPackgNum; i++)
    {
	memset(rxBuffer, 0, DATA_PACKG_SIZE + DATA_PACKG_OVERHEAD);
	
	recv_image_package(rxBuffer, DATA_PACKG_SIZE + DATA_PACKG_OVERHEAD, i);
	dataSize  = (u_int)rxBuffer[2] | ((u_int)rxBuffer[3] << 8);
	//skipping the header		
	pImageBuffer = &rxBuffer[4];

	
	//Start the checksum calculation by adding the first four bytes of the packet
	//TODO: Checksum not working, debug later.
	checkSum = (u_int)rxBuffer[0] + (u_int)rxBuffer[1]
	    + (u_int)rxBuffer[2] + (u_int)rxBuffer[3];
	
	for(j = 0; j < dataSize; j++, pPicBuffer++, pImageBuffer++){
	    *pPicBuffer = *pImageBuffer;
	    checkSum += (u_int)*pImageBuffer;
	}
#ifdef DEBUG
	//Comparing checksum for this frame
	if((char)checkSum != *pImageBuffer)
	{
	    printf("CHECKSUM ERROR\n");
	}
#endif
	
    } /* For */
    
    myPackgIdNum[0] = (char)0xF0;
    myPackgIdNum[1] = (char)0xF0;
    send_ACK(NULL, myPackgIdNum);
    
    save(picBuffer, dataLength, filePath);
    
    return TRUE;
}	

/*********************************** get_dataLenght ***************************/  
int Camera::get_dataLength(char * dataBuffer)
{
    u_int dataLength = 0;

    while(dataBuffer[1] != 0x0a)
    {
#ifdef DEBUG
	printf("get_dataLength: error, Buffer missaligned or wrong data received.\n");
	print_Buffer(dataBuffer);
#endif
    }

    //Concatenating the 3 bytes that determine the length o fthe picture or image in Bytes,
    //Assuming the order of Byte0 = LSB
    dataLength = (u_int) dataBuffer[5] << 16;
    dataLength = dataLength | ((u_int) dataBuffer[4] << 8);
    dataLength = dataLength | ((u_int) dataBuffer[3]);

    return dataLength;
}

/*********************************** recv_command ***************************/  

void Camera::recv_image_package(char * rxBuffer, int rxBufferLen, int counter)
{
    OS_CPU_SR  cpu_sr = 0;
    char myPackgIdNum[2];
    
    //convert the proper packet ID number to 2 bytes
    myPackgIdNum[0] = (char)(counter);
    myPackgIdNum[1] = (char)(counter >> 8);
    at91_usart_receive_frame(&comPort, rxBuffer, rxBufferLen, TIME_OUT);
    at91_usart_trig_cmd (&comPort, US_STTTO);
    send_ACK(NULL, myPackgIdNum);
    OS_ENTER_CRITICAL();
    while (!(at91_usart_get_status(&comPort) & US_TIMEOUT)){/*wait for tx buffer to empty*/}
    OS_EXIT_CRITICAL();
}


/*********************************** send_command ***************************/  

void Camera::send_command(char * txBuffer)
{
    while (!(at91_usart_get_status(&comPort) & US_TXEMPTY)){/*wait for a time out*/}
    at91_usart_send_frame(&comPort, txBuffer, ONE_COMMAND_LEN);
}

/*********************************** send_and_recv ***************************/  
int Camera::send_and_recv(char * txBuffer, char * rxBuffer,
	char ** pRxBuffer, u_int recvByteNum)
{
    OS_CPU_SR  cpu_sr = 0;
    int i;
    
    at91_usart_receive_frame(&comPort, rxBuffer, recvByteNum, TIME_OUT);
    at91_usart_trig_cmd (&comPort, US_STTTO);
    at91_usart_send_frame(&comPort, txBuffer, ONE_COMMAND_LEN);
    OS_ENTER_CRITICAL();

    /* we always get stuck here, fix it */
    //while (!(at91_usart_get_status(&comPort) & US_TIMEOUT)){/*wait for tx buffer to empty*/}

    for(i = 0; i < 100000; i++)
    {
	if (at91_usart_get_status(&comPort) & US_TIMEOUT)
	    break; /* should we do other recovery? */
    }
    
    OS_EXIT_CRITICAL();
    return align_buffer(rxBuffer, pRxBuffer, recvByteNum);
}

/*********************************** print_Buffer ***************************/  
void Camera::print_Buffer(char * rxBuffer)
{
#ifdef DEBUG
    printf("Buffer Contents: 0x%x , 0x%x , 0x%x , 0x%x , 0x%x , 0x%x\n",	
	    rxBuffer[0], rxBuffer[1], rxBuffer[2], 
	    rxBuffer[3], rxBuffer[4], rxBuffer[5]);

#endif
    
}

/*********************************** align_buffer ***************************/      
int Camera::align_buffer(char * dataBuffer, char ** pAlignedBuffer, u_int timeOut)
{
    int i = 0;
    
    while(dataBuffer[i] != 0xAA && timeOut--)
    {
	i++;
    }

    *pAlignedBuffer = &dataBuffer[i];
    return **pAlignedBuffer == 0xAA ? TRUE : FALSE;
}

/*********************************** save ***************************************/
void Camera::save(char * pPicBuffer, int dataLength, char * filePath)
{
    int res;
    int ftp_ctrl_s = FTP_CTRL_SOCKET_CAMERA;
    int ftp_data_s = FTP_DATA_SOCKET_CAMERA;
    
    // save to ftp
    if((res = ftp_client.ftp_open(ftp_ctrl_s, ftp_data_s, filePath, FTP_PASSIVE,
	    WRITE)) == FTP_SUCCESS)
    {
	send_all(ftp_data_s, (uint8*)pPicBuffer, dataLength);	
    }
#ifdef DEBUG
    else
	printf("res = %d\n", res);

#endif
    
    disconnect(ftp_data_s);
    close(ftp_data_s);

    disconnect(ftp_ctrl_s);
    close(ftp_ctrl_s);
}

/*********************************** reset ***************************************/
int Camera::reset()
{
    return send_RESET();
}

/*********************************** powerOff ***************************************/
int Camera::powerOff()
{
    while(!send_POWER_OFF()){};
    
    return TRUE;
}


/*********************************** send_SYNC ***************************/
int Camera::send_SYNC()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[TWO_COMMANDS_LEN + RX_BUFFER_OVERHEAD];
    char * pRxBuffer = NULL;
    int i;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = SYNC;
    txBuffer[2] = SYNC_PARAM1;
    txBuffer[3] = SYNC_PARAM2;
    txBuffer[4] = SYNC_PARAM3;
    txBuffer[5] = SYNC_PARAM4;

    at91_usart_receive_frame(&comPort, rxBuffer, TWO_COMMANDS_LEN + RX_BUFFER_OVERHEAD, TIME_OUT);

    for (i = 0; i < CAMERA_MAX_SYNC_TIMES; i++)
    {
	wait_10ms(5);
	
	at91_usart_send_frame(&comPort, txBuffer, ONE_COMMAND_LEN);

	if (at91_usart_get_status(&comPort) & US_RXRDY)
	{
	    align_buffer(rxBuffer, &pRxBuffer, TWO_COMMANDS_LEN + RX_BUFFER_OVERHEAD);
	    
	    if(is_ack(pRxBuffer,SYNC) && is_sync(&pRxBuffer[ONE_COMMAND_LEN]))
	    {
#ifdef DEBUG
		printf("Camera Connected.\n");

#endif
		send_ACK(SYNC, NULL);
		return TRUE;
	    }
	    else
	    {
		send_NAK(NAK_ERR_6);
		return FALSE;
	    }
	}
    }	
#ifdef DEBUG
    printf("Camera not connected.\n");

#endif
    
    return FALSE;
}


/*********************************** send_ACK ***************************************/
void Camera::send_ACK(u_int commandID, char * packgId)
{
    char txBuffer[ONE_COMMAND_LEN];

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = ACK;
    txBuffer[2] = commandID == NULL ? DATA_RECV : commandID;
    txBuffer[3] = ACK_PARAM2;
    txBuffer[4] = packgId == NULL ? ACK_PARAM3 : packgId[0];
    txBuffer[5] = packgId == NULL ? ACK_PARAM4 : packgId[1];

    send_command(txBuffer);
}

/*********************************** send_NAK ***************************/
void Camera::send_NAK(u_int error)
{
    char txBuffer[ONE_COMMAND_LEN];

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = NAK;
    txBuffer[2] = NAK_PARAM1;
    txBuffer[3] = NAK_PARAM2;
    txBuffer[4] = error;
    txBuffer[5] = NAK_PARAM4;
    
    send_command(txBuffer);
}

/*********************************** send_INITIALIZE ***************************/
int Camera::send_INITIALIZE()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD];

    char * pRxBuffer = NULL;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = INITIALIZE;
    txBuffer[2] = INIT_PARM1;
    txBuffer[3] = COLOR_JPEG;
    txBuffer[4] = PREV_RES_160x120;
    txBuffer[5] = JPEG_RES_320x240;

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD);

    return is_ack(pRxBuffer, INITIALIZE) ? TRUE : FALSE;
}

/*********************************** send_POWER_OFF ***************************************/
int Camera::send_POWER_OFF()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[ONE_COMMAND_LEN + DATA_PACKG_OVERHEAD];
    char * pRxBuffer;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = POWER_OFF;
    txBuffer[2] = POWER_OFF_PARAM1;
    txBuffer[3] = POWER_OFF_PARAM2;
    txBuffer[4] = POWER_OFF_PARAM3;
    txBuffer[5] = POWER_OFF_PARAM4;

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + DATA_PACKG_OVERHEAD);

    return is_ack(pRxBuffer, POWER_OFF) ? TRUE : FALSE;
}

/*********************************** send_RESET ***************************************/
int Camera::send_RESET()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[ONE_COMMAND_LEN + DATA_PACKG_OVERHEAD];
    char * pRxBuffer;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = RESET;
    txBuffer[2] = RST_RESTART;
    txBuffer[3] = RST_PARAM2;
    txBuffer[4] = RST_PARAM3;
    txBuffer[5] = RST_PARAM4;

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + DATA_PACKG_OVERHEAD);

    return is_ack(pRxBuffer, RESET) ? TRUE : FALSE;
}

/*********************************** send_GET_PICTURE **************************/
int Camera::send_GET_PICTURE(int * pDataLength)
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[TWO_COMMANDS_LEN + DATA_PACKG_OVERHEAD];
    char * pRxBuffer;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = GET_PICTURE;
    txBuffer[2] = PIC_JPEG_PREV;
    txBuffer[3] = PIC_PARAM2;
    txBuffer[4] = PIC_PARAM3;
    txBuffer[5] = PIC_PARAM4;

    memset (rxBuffer, 0, TWO_COMMANDS_LEN + DATA_PACKG_OVERHEAD);

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD);


    if(!is_ack(pRxBuffer, GET_PICTURE))
	return FALSE;

    *pDataLength = get_dataLength(&pRxBuffer[6]);

    return TRUE;
}

/***************************** send_SET_PACKG_SIZE ************************/
int Camera::send_SET_PACKG_SIZE()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD];
    char * pRxBuffer;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = SET_PACKG_SIZE;
    txBuffer[2] = SET_PACKG_PARAM1;
    txBuffer[3] = PACKG_SIZE_512_LOW;
    txBuffer[4] = PACKG_SIZE_512_HIGH;
    txBuffer[5] = SET_PACKG_PARAM4;

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD);


    return is_ack(pRxBuffer, SET_PACKG_SIZE) ? TRUE : FALSE;
}

/*********************************** send_SNAPSHOT ***************************/
int Camera::send_SNAPSHOT()
{
    char txBuffer[ONE_COMMAND_LEN];
    char rxBuffer[ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD];
    char * pRxBuffer;

    txBuffer[0] = TRANSMIT;
    txBuffer[1] = SNAPSHOT;
    txBuffer[2] = SNAP_COMPRESS;
    txBuffer[3] = SNAP_SKIP_LOW;
    txBuffer[4] = SNAP_SKIP2_HIGH;
    txBuffer[5] = SNAP_PARAM4;

    send_and_recv(txBuffer, rxBuffer, &pRxBuffer, ONE_COMMAND_LEN + RX_BUFFER_OVERHEAD);

    return is_ack(pRxBuffer, SNAPSHOT) ? TRUE : FALSE;
}

void Camera::takePictures(int num_pics)
{
    INT8U err;
    int i, next;
    char imgpath[32];

    int ftp_ctrl_s = FTP_CTRL_SOCKET_CAMERA;
    int ftp_data_s = FTP_DATA_SOCKET_CAMERA;

    memset(imgpath, 0, 32);

    at91_pio_write (&PIOB_DESC, LED2, LED_ON ) ;

    next = ftp_client.get_last_image_num(ftp_ctrl_s, ftp_data_s, "/SmartHome/www/img/") + 1;
    //next = 0;
	
    OSSemPend(CameraSem, 0, &err);
    
    if (powerOn()){
	for (i = next; i < next + num_pics; i++)
	{
	    sprintf(imgpath, "/SmartHome/www/img/image%03d.jpg", i);
	    get_picture(imgpath);
#ifdef DEBUG
	    printf("took picture %i\n", i);

#endif
	}
	powerOff();
    }

    OSSemPost(CameraSem);

    at91_pio_write (&PIOB_DESC, LED2, LED_OFF ) ;
    
} /*takePictures*/


